@@ -54,6 +54,7 @@ typedef struct _VteTerminalAccessiblePrivate {
 	GArray *snapshot_attributes;	/* Attributes, per byte. */
 	GArray *snapshot_linebreaks;	/* Offsets to line breaks. */
 	gint snapshot_caret;       /* Location of the cursor (in characters). */
+        gboolean text_caret_moved_pending;
 
 	char *action_descriptions[LAST_ACTION];
 } VteTerminalAccessiblePrivate;
@@ -199,7 +200,8 @@ emit_text_changed_delete(GObject *object,
 
 static void
 vte_terminal_accessible_update_private_data_if_needed(VteTerminalAccessible *accessible,
-						      char **old, glong *olen)
+                                                      GString **old_text,
+                                                      GArray **old_characters)
 {
 	VteTerminalAccessiblePrivate *priv = _vte_terminal_accessible_get_instance_private(accessible);
         VteTerminal *terminal;
@@ -212,25 +214,23 @@ vte_terminal_accessible_update_private_data_if_needed(VteTerminalAccessible *acc
 	/* If nothing's changed, just return immediately. */
 	if ((priv->snapshot_contents_invalid == FALSE) &&
 	    (priv->snapshot_caret_invalid == FALSE)) {
-		if (old) {
+                if (old_text) {
 			if (priv->snapshot_text) {
-				*old = g_malloc(priv->snapshot_text->len + 1);
-				memcpy(*old,
-				       priv->snapshot_text->str,
-				       priv->snapshot_text->len);
-				(*old)[priv->snapshot_text->len] = '\0';
-				if (olen) {
-					*olen = priv->snapshot_text->len;
-				}
+                                *old_text = g_string_new_len(priv->snapshot_text->str,
+                                                             priv->snapshot_text->len);
 			} else {
-				*old = g_strdup("");
-				if (olen) {
-					*olen = 0;
-				}
+                                *old_text = g_string_new("");
 			}
-		} else {
-			if (olen) {
-				g_assert_not_reached();
+                }
+                if (old_characters) {
+                        if (priv->snapshot_characters) {
+                                *old_characters = g_array_sized_new(FALSE, FALSE, sizeof(int),
+                                                                    priv->snapshot_characters->len);
+                                g_array_append_vals(*old_characters,
+                                                    priv->snapshot_characters->data,
+                                                    priv->snapshot_characters->len);
+                        } else {
+                                *old_characters = g_array_new(FALSE, FALSE, sizeof(int));
 			}
 		}
 		return;
@@ -241,33 +241,31 @@ vte_terminal_accessible_update_private_data_if_needed(VteTerminalAccessible *acc
 	if (priv->snapshot_contents_invalid) {
 		/* Free the outdated snapshot data, unless the caller
 		 * wants it. */
-		if (old) {
+                if (old_text) {
 			if (priv->snapshot_text != NULL) {
-				*old = priv->snapshot_text->str;
-				if (olen) {
-					*olen = priv->snapshot_text->len;
-				}
-				g_string_free(priv->snapshot_text, FALSE);
+                                *old_text = priv->snapshot_text;
 			} else {
-				*old = g_strdup("");
-				if (olen) {
-					*olen = 0;
-				}
+                                *old_text = g_string_new("");
 			}
 		} else {
-			if (olen) {
-				g_assert_not_reached();
-			}
 			if (priv->snapshot_text != NULL) {
 				g_string_free(priv->snapshot_text, TRUE);
 			}
 		}
 		priv->snapshot_text = NULL;
 
-		/* Free the character offsets and allocate a new array to hold
-		 * them. */
-		if (priv->snapshot_characters != NULL) {
-			g_array_free(priv->snapshot_characters, TRUE);
+                /* Free the character offsets unless the caller wants it,
+                 * and allocate a new array to hold them. */
+                if (old_characters) {
+                        if (priv->snapshot_characters != NULL) {
+                                *old_characters = priv->snapshot_characters;
+                        } else {
+                                *old_characters = g_array_new(FALSE, FALSE, sizeof(int));
+                        }
+                } else {
+                        if (priv->snapshot_characters != NULL) {
+                                g_array_free(priv->snapshot_characters, TRUE);
+                        }
 		}
 		priv->snapshot_characters = g_array_new(FALSE, FALSE, sizeof(int));
 
@@ -344,7 +342,7 @@ vte_terminal_accessible_update_private_data_if_needed(VteTerminalAccessible *acc
 			"Cursor at (%ld, " "%ld).\n", ccol, crow);
 
 	/* Get the offsets to the beginnings of each line. */
-	caret = -1;
+	caret = 0;
 	for (i = 0; i < priv->snapshot_characters->len; i++) {
 		/* Get the attributes for the current cell. */
 		offset = g_array_index(priv->snapshot_characters,
@@ -360,16 +358,11 @@ vte_terminal_accessible_update_private_data_if_needed(VteTerminalAccessible *acc
 		}
 	}
 
-	/* If no cells are before the caret, then the caret must be
-	 * at the end of the buffer. */
-	if (caret == -1) {
-		caret = priv->snapshot_characters->len;
-	}
-
-	/* Notify observers if the caret moved. */
+        /* Make a note that we'll need to notify observers if the caret moved.
+         * But only notify them after sending text-changed. */
 	if (caret != priv->snapshot_caret) {
 		priv->snapshot_caret = caret;
-		emit_text_caret_moved(G_OBJECT(accessible), caret);
+                priv->text_caret_moved_pending = TRUE;
 	}
 
 	/* Done updating the caret position, whether we needed to or not. */
@@ -382,12 +375,25 @@ vte_terminal_accessible_update_private_data_if_needed(VteTerminalAccessible *acc
 			(long)priv->snapshot_characters->len);
 }
 
+static void
+vte_terminal_accessible_maybe_emit_text_caret_moved(VteTerminalAccessible *accessible)
+{
+        VteTerminalAccessiblePrivate *priv = _vte_terminal_accessible_get_instance_private(accessible);
+
+        if (priv->text_caret_moved_pending) {
+                emit_text_caret_moved(G_OBJECT(accessible), priv->snapshot_caret);
+                priv->text_caret_moved_pending = FALSE;
+        }
+}
+
 /* A signal handler to catch "text-inserted/deleted/modified" signals. */
 static void
 vte_terminal_accessible_text_modified(VteTerminal *terminal, gpointer data)
 {
         VteTerminalAccessible *accessible = data;
 	VteTerminalAccessiblePrivate *priv = _vte_terminal_accessible_get_instance_private(accessible);
+        GString *old_text;
+        GArray *old_characters;
 	char *old, *current;
 	glong offset, caret_offset, olen, clen;
 	gint old_snapshot_caret;
@@ -395,11 +401,15 @@ vte_terminal_accessible_text_modified(VteTerminal *terminal, gpointer data)
 	old_snapshot_caret = priv->snapshot_caret;
 	priv->snapshot_contents_invalid = TRUE;
 	vte_terminal_accessible_update_private_data_if_needed(accessible,
-							      &old, &olen);
-	g_assert(old != NULL);
+                                                              &old_text,
+                                                              &old_characters);
+        g_assert(old_text != NULL);
+        g_assert(old_characters != NULL);
 
 	current = priv->snapshot_text->str;
 	clen = priv->snapshot_text->len;
+        old = old_text->str;
+        olen = old_text->len;
 
 	if ((guint) priv->snapshot_caret < priv->snapshot_characters->len) {
 		caret_offset = g_array_index(priv->snapshot_characters,
@@ -422,12 +432,15 @@ vte_terminal_accessible_text_modified(VteTerminal *terminal, gpointer data)
 	if ((olen == offset) &&
 		       	(caret_offset < olen && old[caret_offset] == ' ') &&
 			(old_snapshot_caret == priv->snapshot_caret + 1)) {
-                priv->snapshot_text->str = old;
-		priv->snapshot_text->len = caret_offset + 1;
+                GString *saved_text = priv->snapshot_text;
+                GArray *saved_characters = priv->snapshot_characters;
+
+                priv->snapshot_text = old_text;
+                priv->snapshot_characters = old_characters;
 		emit_text_changed_delete(G_OBJECT(accessible),
 					 old, caret_offset, 1);
-		priv->snapshot_text->str = current;
-		priv->snapshot_text->len = clen;
+                priv->snapshot_text = saved_text;
+                priv->snapshot_characters = saved_characters;
 	}
 
 
@@ -456,17 +469,17 @@ vte_terminal_accessible_text_modified(VteTerminal *terminal, gpointer data)
 		/* Now emit a deleted signal for text that was in the old
 		 * string but isn't in the new one... */
 		if (olen > offset) {
-			gchar *saved_str = priv->snapshot_text->str;
-			gsize saved_len = priv->snapshot_text->len;
+                        GString *saved_text = priv->snapshot_text;
+                        GArray *saved_characters = priv->snapshot_characters;
 
-			priv->snapshot_text->str = old;
-			priv->snapshot_text->len = olen;
+                        priv->snapshot_text = old_text;
+                        priv->snapshot_characters = old_characters;
 			emit_text_changed_delete(G_OBJECT(accessible),
 						 old,
 						 offset,
 						 olen - offset);
-			priv->snapshot_text->str = saved_str;
-			priv->snapshot_text->len = saved_len;
+                        priv->snapshot_text = saved_text;
+                        priv->snapshot_characters = saved_characters;
 		}
 		/* .. and an inserted signal for text that wasn't in the old
 		 * string but is in the new one. */
@@ -478,7 +491,10 @@ vte_terminal_accessible_text_modified(VteTerminal *terminal, gpointer data)
 		}
 	}
 
-	g_free(old);
+        vte_terminal_accessible_maybe_emit_text_caret_moved(accessible);
+
+        g_string_free(old_text, TRUE);
+        g_array_free(old_characters, TRUE);
 }
 
 /* A signal handler to catch "text-scrolled" signals. */
@@ -520,6 +536,7 @@ vte_terminal_accessible_text_scrolled(VteTerminal *terminal,
 							 priv->snapshot_text->len);
 			}
 		}
+                vte_terminal_accessible_maybe_emit_text_caret_moved(accessible);
 		return;
 	}
 	/* Find the start point. */
@@ -574,6 +591,7 @@ vte_terminal_accessible_text_scrolled(VteTerminal *terminal,
 							 len - i);
 			}
 		}
+                vte_terminal_accessible_maybe_emit_text_caret_moved(accessible);
 		return;
 	}
 	/* We scrolled down, so text was added at the bottom and removed
@@ -609,12 +627,16 @@ vte_terminal_accessible_text_scrolled(VteTerminal *terminal,
 		if (inserted) {
 			len = priv->snapshot_text->len;
 			if (len > i) {
+				/* snapshot_text always contains a trailing '\n',
+				 * insertion happens in front of it: bug 657960 */
+				g_assert(i >= 1);
 				emit_text_changed_insert(G_OBJECT(accessible),
 							 priv->snapshot_text->str,
-							 i,
+							 i - 1,
 							 len - i);
 			}
 		}
+                vte_terminal_accessible_maybe_emit_text_caret_moved(accessible);
 		return;
 	}
 	g_assert_not_reached();
@@ -632,6 +654,7 @@ vte_terminal_accessible_invalidate_cursor(VteTerminal *terminal, gpointer data)
 	priv->snapshot_caret_invalid = TRUE;
 	vte_terminal_accessible_update_private_data_if_needed(accessible,
 							      NULL, NULL);
+        vte_terminal_accessible_maybe_emit_text_caret_moved(accessible);
 }
 
 /* Handle title changes by resetting the description. */
@@ -748,6 +771,7 @@ _vte_terminal_accessible_init (VteTerminalAccessible *accessible)
 	priv->snapshot_caret = -1;
 	priv->snapshot_contents_invalid = TRUE;
 	priv->snapshot_caret_invalid = TRUE;
+        priv->text_caret_moved_pending = FALSE;
 }
 
 static void
@@ -883,12 +907,15 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 {
         VteTerminalAccessible *accessible = VTE_TERMINAL_ACCESSIBLE(text);
 	VteTerminalAccessiblePrivate *priv = _vte_terminal_accessible_get_instance_private(accessible);
+        VteTerminal *terminal;
 	gunichar current, prev, next;
 	guint start, end, line;
 
 	vte_terminal_accessible_update_private_data_if_needed(accessible,
 							      NULL, NULL);
 
+        terminal = VTE_TERMINAL(gtk_accessible_get_widget (GTK_ACCESSIBLE(text)));
+
 	_vte_debug_print(VTE_DEBUG_ALLY,
 			"Getting %s %s at %d of %d.\n",
 			(direction == direction_current) ? "this" :
@@ -921,7 +948,7 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 			/* Back up to the previous non-word-word transition. */
 			while (offset > 0) {
 				prev = vte_terminal_accessible_get_character_at_offset(text, offset - 1);
-				if (_vte_is_word_char(prev)) {
+				if (_vte_terminal_is_word_char(terminal, prev)) {
 					offset--;
 				} else {
 					break;
@@ -935,7 +962,7 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 			if (direction == direction_previous) {
 				while (offset > 0) {
 					prev = vte_terminal_accessible_get_character_at_offset(text, offset - 1);
-					if (!_vte_is_word_char(prev)) {
+					if (!_vte_terminal_is_word_char(terminal, prev)) {
 						offset--;
 					} else {
 						break;
@@ -943,7 +970,7 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 				}
 				while (offset > 0) {
 					prev = vte_terminal_accessible_get_character_at_offset(text, offset - 1);
-					if (_vte_is_word_char(prev)) {
+					if (_vte_terminal_is_word_char(terminal, prev)) {
 						offset--;
 					} else {
 						break;
@@ -958,7 +985,7 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 			if (direction == direction_next) {
 				while (offset < (int) priv->snapshot_characters->len) {
 					next = vte_terminal_accessible_get_character_at_offset(text, offset);
-					if (_vte_is_word_char(next)) {
+					if (_vte_terminal_is_word_char(terminal, next)) {
 						offset++;
 					} else {
 						break;
@@ -966,7 +993,7 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 				}
 				while (offset < (int) priv->snapshot_characters->len) {
 					next = vte_terminal_accessible_get_character_at_offset(text, offset);
-					if (!_vte_is_word_char(next)) {
+					if (!_vte_terminal_is_word_char(terminal, next)) {
 						offset++;
 					} else {
 						break;
@@ -977,7 +1004,7 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 			/* Now find the end of this word. */
 			while (offset < (int) priv->snapshot_characters->len) {
 				current = vte_terminal_accessible_get_character_at_offset(text, offset);
-				if (_vte_is_word_char(current)) {
+				if (_vte_terminal_is_word_char(terminal, current)) {
 					offset++;
 				} else {
 					break;
@@ -987,7 +1014,7 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 			/* Now find the next non-word-word transition */
 			while (offset < (int) priv->snapshot_characters->len) {
 				next = vte_terminal_accessible_get_character_at_offset(text, offset);
-				if (!_vte_is_word_char(next)) {
+				if (!_vte_terminal_is_word_char(terminal, next)) {
 					offset++;
 				} else {
 					break;
@@ -1000,8 +1027,8 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 			current = vte_terminal_accessible_get_character_at_offset(text, offset);
 			while (offset > 0) {
 				prev = vte_terminal_accessible_get_character_at_offset(text, offset - 1);
-				if (_vte_is_word_char(prev) &&
-				    !_vte_is_word_char(current)) {
+				if (_vte_terminal_is_word_char(terminal, prev) &&
+				    !_vte_terminal_is_word_char(terminal, current)) {
 					break;
 				} else {
 					offset--;
@@ -1016,7 +1043,7 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 			if (direction == direction_previous) {
 				while (offset > 0) {
 					prev = vte_terminal_accessible_get_character_at_offset(text, offset - 1);
-					if (_vte_is_word_char(prev)) {
+					if (_vte_terminal_is_word_char(terminal, prev)) {
 						offset--;
 					} else {
 						break;
@@ -1025,8 +1052,8 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 				current = vte_terminal_accessible_get_character_at_offset(text, offset);
 				while (offset > 0) {
 					prev = vte_terminal_accessible_get_character_at_offset(text, offset - 1);
-					if (_vte_is_word_char(prev) &&
-					    !_vte_is_word_char(current)) {
+					if (_vte_terminal_is_word_char(terminal, prev) &&
+					    !_vte_terminal_is_word_char(terminal, current)) {
 						break;
 					} else {
 						offset--;
@@ -1042,7 +1069,7 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 			if (direction == direction_next) {
 				while (offset < (int) priv->snapshot_characters->len) {
 					current = vte_terminal_accessible_get_character_at_offset(text, offset);
-					if (!_vte_is_word_char(current)) {
+					if (!_vte_terminal_is_word_char(terminal, current)) {
 						offset++;
 					} else {
 						break;
@@ -1050,7 +1077,7 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 				}
 				while (offset < (int) priv->snapshot_characters->len) {
 					current = vte_terminal_accessible_get_character_at_offset(text, offset);
-					if (_vte_is_word_char(current)) {
+					if (_vte_terminal_is_word_char(terminal, current)) {
 						offset++;
 					} else {
 						break;
@@ -1061,7 +1088,7 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 			/* Now find the next word end. */
 			while (offset < (int) priv->snapshot_characters->len) {
 				current = vte_terminal_accessible_get_character_at_offset(text, offset);
-				if (!_vte_is_word_char(current)) {
+				if (!_vte_terminal_is_word_char(terminal, current)) {
 					offset++;
 				} else {
 					break;
@@ -1069,7 +1096,7 @@ vte_terminal_accessible_get_text_somewhere(AtkText *text,
 			}
 			while (offset < (int) priv->snapshot_characters->len) {
 				current = vte_terminal_accessible_get_character_at_offset(text, offset);
-				if (_vte_is_word_char(current)) {
+				if (_vte_terminal_is_word_char(terminal, current)) {
 					offset++;
 				} else {
 					break;
