@@ -172,7 +172,7 @@ void vte_terminal_unselect_all(VteTerminal *terminal) _VTE_GNUC_NONNULL(1);
 
 /* By-word selection */
 void vte_terminal_set_word_char_exceptions(VteTerminal *terminal,
-                                           const char *word_char_exceptions) _VTE_GNUC_NONNULL(1);
+                                           const char *exceptions) _VTE_GNUC_NONNULL(1);
 const char *vte_terminal_get_word_char_exceptions(VteTerminal *terminal) _VTE_GNUC_NONNULL(1);
 
 /* Set the terminal's size. */
